
; Convert a Sound with ACM
; by Zapman from a C code found at http://david.weekly.org and with some tips from the CDex code of Albert L. Faber
; Not yet tested on MPEG 1,2 and 2.5

IncludeFile ("ACMConstAndStruct.pb")
IncludeFile ("MP3_Tags.pb")

Procedure.s TextError () 
; by Zapman
; When an error occure, this fonction retrieve the text of it
; Quand une erreur survient, cette fonction rcupre le text correspondant
  errorNb = GetLastError_ () 
  If errorNb 
    hmem = AllocateMemory ( 255)
    Length = FormatMessage_ (#FORMAT_MESSAGE_FROM_SYSTEM, #Null, errorNb, 0, hmem, 255, #Null) 
    If Length > 1 
        e$ = PeekS (hmem, Length - 2) 
    Else 
        e$ = "Erreur N / Error N# "+Str(errorNb) 
    EndIf 
    FreeMemory (hmem) 
    ProcedureReturn e$ 
  Else 
    ProcedureReturn "" 
  EndIf 
EndProcedure
;
Procedure ShowError()
; by Zapman
  Err$=TextError()
  If Err$<>""
    MessageRequester("Error", Err$, 0)
  EndIf
EndProcedure


Procedure DebugACMError(e.l,FirstLine$)
  tx.s = ""
  Select e
  Case #MMSYSERR_INVALFLAG
    tx = "INVALIDE FLAG"
  Case #MMSYSERR_INVALHANDLE
    tx = "INVALIDE HANDLE"
  Case #MMSYSERR_INVALPARAM
    tx = "INVALIDE PARAM"
  Case #MMSYSERR_NODRIVER
    tx = "NO DRIVER"
    MessageRequester("Error",FirstLine$+Chr(13)+"Sorry, there is not any driver available for this format.",0)
  Case #ACMERR_NOTPOSSIBLE
    tx = "The operation is not possible or the details requested are not available."
  EndSelect
  If tx And tx<>"NO DRIVER"
    MessageRequester("Error",FirstLine$+Chr(13)+"Unexpected error: "+tx+Chr(13)+"Please contact the author to report that error.",0)
  EndIf
EndProcedure

;
;find the biggest format size
;
maxFormatSize = 0;
acmMetrics_(0, #ACM_METRIC_MAX_SIZE_FORMAT,@maxFormatSize)

; --------------------------------------
;
FileName$ = OpenFileRequester("Open a sound file", "", "*.*", 0)
If FileName$
  If OpenFile(1, FileName$)
    block_size_in = 512
    RError = 0
    *InputArea = AllocateMemory(maxFormatSize) ; Allocate room for input data definition
    RtlZeroMemory_(*InputArea, maxFormatSize)  ; Clean the input data area
    If LCase(Right(FileName$,3))="wav"
      *hmem=AllocateMemory(36)
      ReadData(1,*hmem,36)
      *WInputFormat.WAVEFORMATEX = *InputArea
      CopyMemory(*hmem+20,*WInputFormat,16) ; copy nChannels, nSamplesPerSec, wBitsPerSample, nBlockAlign and nAvgBytesPerSec from the open file
      block_size_in = 512**WInputFormat\nBlockAlign ; it could be 1024 or anything else instead of 512
    EndIf
    If LCase(Right(FileName$,3))="mp3" Or LCase(Right(FileName$,3))="mpg"
      TagString$ = DecodeMP3Tags(0)
      If TagString$<>"Error"
        ;
        ; define mpeg input format
        ;
        MpgVersion.s = ReturnTagValue ("MpgVersion",TagString$)
        Layer = Val(ReturnTagValue ("Layer",TagString$))
        nBitRate = Val(ReturnTagValue ("BitRate",TagString$))
        
        *WInputFormat.WAVEFORMATEX = *InputArea
        *WInputFormat\nChannels = Val(ReturnTagValue ("Channels",TagString$))
        *WInputFormat\nSamplesPerSec = Val(ReturnTagValue ("SampleRate",TagString$));  Khtz
        
        
        *WInputFormat\nAvgBytesPerSec = nBitRate * 1000 / 8;	// nBitrate in bytes per sec 
        *WInputFormat\nBlockAlign  = 1;					// assume variable bit rate     
        *WInputFormat\wBitsPerSample = 0;                         // not used
        
          
        If Layer = 3
          *MP3InputFormat.MPEGLAYER3WAVEFORMAT = *InputArea
          *MP3InputFormat\wfx\wFormatTag = #WAVE_FORMAT_MPEGLAYER3
          *MP3InputFormat\wfx\cbSize   = #MPEGLAYER3_WFX_EXTRA_BYTES;	      // default
          *MP3InputFormat\wID     = #MPEGLAYER3_ID_MPEG;
          *MP3InputFormat\fdwFlags    = #MPEGLAYER3_FLAG_PADDING_OFF;
          *MP3InputFormat\nBlockSize   = Val(ReturnTagValue ("FrameLength",TagString$))
          *MP3InputFormat\nFramesPerBlock  = $0001;
          *MP3InputFormat\nCodecDelay   = $0571;
          
          ; The following comes from Albert L. Faber, I'm not sure that it is right
          ; --------------------------------------------------------------
          Select nBitRate
            Case 18: *MP3InputFormat\nFramesPerBlock = $0004
            Case 20: *MP3InputFormat\nFramesPerBlock = $0002
          EndSelect
          ; --------------------------------------------------------------
          
          block_size_in = *MP3InputFormat\nBlockSize
          
        ElseIf Layer = 1
          ; This (MPEG 1 and 2) is not yet tested
          *MP1InputFormat.MPEG1WAVEFORMAT = *InputArea
          *MP1InputFormat\wfx\wFormatTag = #WAVE_FORMAT_MPEG
          *MP1InputFormat\wfx\cbSize   = #MPEG_WFX_EXTRA_BYTES
          *MP1InputFormat\fwHeadLayer = #ACM_MPEG_LAYER1
          *MP1InputFormat\dwHeadBitrate = nBitRate *1000
          *MP1InputFormat\fwHeadMode = Val(ReturnTagValue ("Channels",TagString$))
          *MP1InputFormat\fwHeadModeExt = 0
          *MP1InputFormat\wHeadEmphasis = Val(ReturnTagValue ("Emphasis",TagString$))
          *MP1InputFormat\fwHeadFlags = #ACM_MPEG_ID_MPEG1
          *MP1InputFormat\dwPTSLow = 0
          *MP1InputFormat\dwPTSHigh = 0
          
        ElseIf Layer = 2
          ; This (MPEG 1 and 2) is not yet tested
          *MP2InputFormat.MPEG1WAVEFORMAT = *InputArea
          *MP2InputFormat\wfx\wFormatTag = #WAVE_FORMAT_MPEG;			// Set format tag
          *MP2InputFormat\wfx\cbSize = #MPEG_WFX_EXTRA_BYTES;	            // Set number of extra bytes
          *MP2InputFormat\fwHeadLayer = #ACM_MPEG_LAYER2
          *MP2InputFormat\dwHeadBitrate = nBitRate*1000;				// Bits per second
          *MP2InputFormat\fwHeadMode = Val(ReturnTagValue ("Channels",TagString$))
              ;ACM_MPEG_STEREO:ACM_MPEG_SINGLECHANNEL;		// Set mode (MONO/JOINT/STEREO/DUAL)
          *MP2InputFormat\fwHeadModeExt = 0;							// Extra mode information for joint stereo
          *MP2InputFormat\wHeadEmphasis = Val(ReturnTagValue ("Emphasis",TagString$));							// No emphasis
          *MP2InputFormat\fwHeadFlags = $14;	ACM_MPEG_ID_MPEG2 ??                // MPEG flags (optionally can set more information like private etc
          *MP2InputFormat\dwPTSLow = 0;								// Not associated with video stream
          *MP2InputFormat\dwPTSHigh = 0;								// Not associated with video stream
        
        EndIf

      Else
        RError = 1
      EndIf
    EndIf
    CloseFile(1)
    If RError
      MessageRequester("Error","Unknown format.",0)
    Else
      ;
      ; Initialize the ACMFORMATCHOOSE members.
      ;
      *OutputFormat = AllocateMemory(maxFormatSize)
      afc.ACMFORMATCHOOSE\cbStruct    = SizeOf(ACMFORMATCHOOSE);
      afc\fdwStyle    = 0;            no special style flags
      afc\hwndOwner   = 0;            hwnd of parent window
      afc\pwfx        = *OutputFormat
      afc\cbwfx       = maxFormatSize
      afc\pszTitle    = @"Select the new format"
      afc\fdwEnum =     #ACM_FORMATENUMF_CONVERT
      afc\pwfxEnum = *InputArea
      ;
      
      mmr = acmFormatChoose_(@afc)
      If mmr
        tx$ = "acmFormatChoose didn't success ! "
        DebugACMError(mmr,"acmFormatChoose didn't success ! ")
      Else 

        ConversionStream = 0
        mmr=acmStreamOpen_(@ConversionStream,#Null,*InputArea,*OutputFormat,#Null,0,0,#ACM_STREAMOPENF_NONREALTIME)
        If mmr
          DebugACMError(mmr,"acmStreamOpen didn't success ! ")
        Else
          ;
          ;Stream converter opened correctly
          ;now, let's open a file, read in a bunch of data, and convert it!
          ;
          If OpenFile(1, FileName$) = 0
            Debug "Error opening the input file"
            ShowError()
          Else
            DataInTotSize = 0
            *WInputFormat.WAVEFORMATEX = *InputArea
            If *WInputFormat\wFormatTag = #WAVE_FORMAT_PCM ; source
              posInFile = 16
              tx$=ReadString(1)
              While Left(tx$,4)<>"data" And posInFile <Lof(1) ; look for the data chunck
                FileSeek(1,posInFile)
                ReadData(1,@chunckLenght.l,4)
                posInFile = posInFile + chunckLenght + 4
                FileSeek(1,posInFile)
                tx$=ReadString(1)
              Wend
              posInFile + 4
              FileSeek(1,posInFile)
              ReadData(1,@DataInTotSize,4)
              posInFile + 4
              FileSeek(1,posInFile)
            EndIf
            ;
            ; find out how big the decoded buffer will be
            ;
            block_size_out = 0;
            mmr = acmStreamSize_(ConversionStream, block_size_in, @block_size_out, #ACM_STREAMSIZEF_SOURCE)
            If mmr<>0 Or block_size_out=0
              DebugACMError(mmr,"acmStreamSize didn't success ! ")
            Else 
              ;
              ; allocate our I/O buffers
              ;
              Sourcebuf = AllocateMemory(block_size_in)
              Destbuf   = AllocateMemory(block_size_out)
              ;
              ; prepare the decoder
              ;
              StreamHead.ACMSTREAMHEADER
              RtlZeroMemory_(@StreamHead, SizeOf(ACMSTREAMHEADER))
              StreamHead\cbStruct = SizeOf(ACMSTREAMHEADER )
              StreamHead\pbSrc = Sourcebuf
              StreamHead\cbSrcLength = block_size_in
              StreamHead\pbDst = Destbuf 
              StreamHead\cbDstLength = block_size_out
              mmr=0
              mmr = acmStreamPrepareHeader_(ConversionStream, @StreamHead, 0)
              If mmr = 0
                ;
                ; Dump this data off to disk
                ;
                While Right(FileName$,1)<>"."
                  FileName$=Left(FileName$,Len(FileName$)-1)
                Wend
                FileName$=Left(FileName$,Len(FileName$)-1)
                *WOutputFormat.WAVEFORMATEX = *OutputFormat
                Select *WOutputFormat\wFormatTag 
                  Case #WAVE_FORMAT_PCM 
                    FileName$+"2.wav"
                  Case #WAVE_FORMAT_MPEGLAYER3
                    FileName$+"2.mp3"
                  Case #WAVE_FORMAT_WMA2
                    FileName$+"2.wma"
                  Case #WAVE_FORMAT_ALAW
                    FileName$+"2.AIFF"
                  Case #WAVE_FORMAT_MULAW
                    FileName$+"2.AIFF"
                  Default
                    FileName$+"2.___"
                EndSelect
                If FileSize(FileName$)>0 ; if the file yet exists, delete it
                  DeleteFile(FileName$)
                EndIf
                If  OpenFile(2, FileName$) = 0
                  MessageRequester("Error","Error while creating the output file!!",0)
                Else
                  If *WOutputFormat\wFormatTag = #WAVE_FORMAT_PCM 
                    ;
                    ; Generate the PCM header
                    ;
                    Hdata$="RIFF"
                    WriteData (2,@Hdata$,4)
                    NData.l=0 ; Total size (this will be updated later)
                    WriteData (2,@NData,4)
                    Hdata$="WAVEfmt "
                    WriteData (2,@Hdata$,8)
                    NData.l=16 ; Header size
                    WriteData (2,@NData,4)
                    WriteData (2,*OutputFormat,16)
                    Hdata$="data"
                    WriteData (2,@Hdata$,4)
                    NData.l=0
                    WriteData (2,@NData,4) ; datas size (this will be updated later)
                  EndIf
                  DataOutSize = 0
                  DataInSize = 0
                  cont = 1
                  ;UseFile(1)
                  OpenWindow(0,0,0,200,30,"Processing...",#PB_Window_ScreenCentered)
                  TextGadget(500, 4, 5, 190, 20,"Loading file...",#PB_Text_Center)
                  Sometimes = 0
                  While cont
                    Sometimes + 1
                    If Sometimes = 100
                      Sometimes  = 0
                      SetGadgetText(500, "Input: "+Str(DataInSize)+"  -  Output: "+Str(DataOutSize))
                    EndIf
                    ;
                    ; Suck in some Source Data
                    ;
                    ;UseFile(1)
                    If DataInTotSize 
                      If DataInSize + block_size_in > DataInTotSize 
                        block_size_in = DataInTotSize - DataInSize
                        StreamHead\cbSrcLength = block_size_in
                      EndIf
                    EndIf
                    count = ReadData(1,Sourcebuf,block_size_in)
                    DataInSize + count
                    If DataInTotSize And DataInSize = DataInTotSize
                      cont = 0
                    EndIf 
                    If (count <> block_size_in)
                      cont = 0
                    EndIf
                    ;
                    ; Convert the Data
                    ;
                    mmr = acmStreamConvert_(ConversionStream,@StreamHead, 0);#ACM_STREAMCONVERTF_BLOCKALIGN)
                    
                    ;
                    ; Write the decoded datas To disk
                    ;
                    If mmr=0 And StreamHead\cbDstLengthUsed
                      ;UseFile (2)
                      WriteData (2,Destbuf , StreamHead\cbDstLengthUsed)
                      DataOutSize + StreamHead\cbDstLengthUsed
                    Else
                      DebugACMError(mmr,"acmStreamConvert didn't success!")
                    EndIf 
                  Wend
                  CloseWindow(0)
                  ;
                  If *WOutputFormat\wFormatTag = #WAVE_FORMAT_PCM ; Update sizes
                    ;UseFile (2)
                    FileSeek(2,4)
                    NData=DataOutSize + 44
                    WriteData (2,@NData,4) ; Update total size
                    FileSeek(2,40)
                    NData=DataOutSize 
                    WriteData (2,@NData,4) ; Update datas size
                  EndIf
                  CloseFile(2)
                EndIf
                ;
                ; Clean up after yourself like a good little boy
                ;
                acmStreamUnprepareHeader_(ConversionStream, @StreamHead,0)
              Else
                DebugACMError(mmr,"acmStreamPrepareHeader didn't success ! ")
              EndIf
              FreeMemory(Destbuf)
              FreeMemory(Sourcebuf)
            EndIf 
            CloseFile(1)
          EndIf
          acmStreamClose_(ConversionStream, 0)
        EndIf
      EndIf
      FreeMemory(*OutputFormat)
    EndIf
    FreeMemory(*InputArea)
  EndIf
EndIf

End
; IDE Options = PureBasic 5.24 LTS (Windows - x86)
; CursorPosition = 286
; FirstLine = 262
; Folding = -